use crate::prelude::*;
use anyhow::Result;
use ecs::{Resource, resources::Resource as ResourceTrait};
use std::collections::HashMap;

#[derive(Debug)]
pub enum EngineEvent<'a> {
    MouseMotion(u32, u32),
    MouseButtonDown(MouseButton),
    MouseButtonUp(MouseButton),
    MouseWheel(i32, i32, MouseWheelDirection),

    KeyDown(Keycode),
    KeyUp(Keycode),

    ControllerButtonDown(&'a Controller, ControllerButton),
    ControllerButtonUp(&'a Controller, ControllerButton),
    ControllerAxis(&'a Controller, ControllerAxis),
    ControllerAdded(&'a Controller),
    ControllerRemoved(&'a Controller),

    JoystickButtonDown(&'a Joystick),
    JoystickButtonUp(&'a Joystick),
    JoystickAxis(&'a Joystick, u8, i16),
    JoystickAdded(&'a Joystick),
    JoystickRemoved(&'a Joystick),

    FileDrop(String),
}

#[derive(Resource)]
pub struct InputMap {
    pub direction_mapping: HashMap<Keycode, GuiDirection>,
}

pub trait EventSystem<Func, T> {
    fn event_system(world_builder: &mut WorldBuilder, func: Func);
}

impl<Func, T> EventSystem<Func, T> for Engine
where
    Func: Fn(&mut T, &mut Commands, EngineEvent<'_>) -> Result<()> + Send + Sync + 'static,
    T: ResourceTrait,
{
    fn event_system(world_builder: &mut WorldBuilder, func: Func) {
        world_builder.add_system(
            0,
            |world: &World,
             commands: &mut Commands,
             consumer: &mut T,
             gui_handler: &mut GuiHandler,
             input_map: &mut InputMap,
             context: &mut Context|
             -> Result<bool> {
                let res = context.events(world, |event| {
                    Self::event(commands, gui_handler, input_map, consumer, event, func)
                })?;

                gui_handler.process_callbacks(commands)?;

                Ok(res)
            },
        )
    }
}

impl Engine {
    pub(crate) fn event<T>(
        commands: &mut Commands,
        gui_handler: &mut GuiHandler,
        input: &InputMap,
        consumer: &mut T,
        event: Event<'_>,
        func: Func,
    ) -> anyhow::Result<()> {
        match event {
            Event::MouseMotion(x, y) => {
                gui_handler.set_mouse_pos(x, y)?;
                consumer.event(commands, EngineEvent::MouseMotion(x, y))?;
            }
            Event::MouseButtonDown(mouse_button) => {
                if !gui_handler.mouse_down(mouse_button)? {
                    consumer.event(commands, EngineEvent::MouseButtonDown(mouse_button))?;
                }
            }
            Event::MouseButtonUp(mouse_button) => {
                if !gui_handler.mouse_up(mouse_button)? {
                    consumer.event(commands, EngineEvent::MouseButtonUp(mouse_button))?;
                }
            }
            Event::MouseWheel(x, y, direction) => {
                consumer.event(commands, EngineEvent::MouseWheel(x, y, direction))?
            }
            Event::KeyDown(keycode) => {
                Self::key_down_event(commands, gui_handler, input, consumer, keycode)?;
            }
            Event::KeyUp(keycode) => {
                Self::key_up_event(commands, gui_handler, input, consumer, keycode)?;
            }
            Event::TextInput(text) => {
                Self::text_input(gui_handler, text)?;
            }
            Event::ControllerButtonDown(controller, button) => {
                Self::button_down_event(commands, gui_handler, consumer, controller, button)?;
            }
            Event::ControllerButtonUp(controller, button) => {
                Self::button_up_event(commands, consumer, controller, button)?;
            }
            Event::ControllerAxis(controller) => {
                if !gui_handler.check_navigatable() {
                    Self::axis_event(commands, consumer, &controller)?
                } else {
                    gui_handler.update_selection(controller.direction())?;
                }
            }
            Event::ControllerAdded(controller) => {
                Self::controller_added(commands, consumer, controller)?
            }
            Event::ControllerRemoved(controller) => {
                Self::controller_removed(commands, consumer, controller)?
            }

            Event::JoystickAxis(joystick, axis_index, value) => consumer.event(
                commands,
                EngineEvent::JoystickAxis(joystick, axis_index, value),
            )?,
            Event::JoystickButtonDown(joystick) => {
                consumer.event(commands, EngineEvent::JoystickButtonDown(joystick))?
            }
            Event::JoystickButtonUp(joystick) => {
                consumer.event(commands, EngineEvent::JoystickButtonUp(joystick))?
            }
            Event::JoystickAdded(joystick) => {
                consumer.event(commands, EngineEvent::JoystickAdded(joystick))?
            }
            Event::JoystickRemoved(joystick) => {
                consumer.event(commands, EngineEvent::JoystickRemoved(joystick))?
            }

            Event::FileDrop(filename) => {
                consumer.event(commands, EngineEvent::FileDrop(filename))?
            }
        }

        Ok(())
    }
}

impl Engine {
    #[inline]
    fn controller_added<T: EventConsumer>(
        commands: &mut Commands,
        consumer: &mut T,
        controller: &Controller,
    ) -> Result<()> {
        consumer.event(commands, EngineEvent::ControllerAdded(controller))?;

        Ok(())
    }

    #[inline]
    fn controller_removed<T: EventConsumer>(
        commands: &mut Commands,
        consumer: &mut T,
        controller: &Controller,
    ) -> Result<()> {
        consumer.event(commands, EngineEvent::ControllerRemoved(controller))?;

        Ok(())
    }

    #[inline]
    fn key_up_event<T: EventConsumer>(
        commands: &mut Commands,
        gui_handler: &mut GuiHandler,
        input: &InputMap,
        consumer: &mut T,
        keycode: Keycode,
    ) -> Result<()> {
        if input.direction_mapping.get(&keycode).is_some()
            && gui_handler.update_selection(GuiDirection::None)?
        {
            return Ok(());
        }

        consumer.event(commands, EngineEvent::KeyUp(keycode))?;

        Ok(())
    }

    #[inline]
    fn key_down_event<T: EventConsumer>(
        commands: &mut Commands,
        gui_handler: &mut GuiHandler,
        input: &InputMap,
        consumer: &mut T,
        keycode: Keycode,
    ) -> Result<()> {
        if let Some(direction) = input.direction_mapping.get(&keycode) {
            if gui_handler.update_selection(*direction)? {
                return Ok(());
            }
        }

        match keycode {
            Keycode::Backspace => {
                if gui_handler.remove_char()? {
                    return Ok(());
                }
            }
            Keycode::Return => {
                if gui_handler.accept_selection()? {
                    return Ok(());
                }
            }
            Keycode::Escape => {
                if gui_handler.decline_topgui(commands)? {
                    return Ok(());
                }
            }

            _ => (),
        }

        consumer.event(commands, EngineEvent::KeyDown(keycode))?;

        Ok(())
    }

    #[inline]
    fn button_up_event<T: EventConsumer>(
        commands: &mut Commands,
        consumer: &mut T,
        controller: &Controller,
        button: ControllerButton,
    ) -> Result<()> {
        consumer.event(
            commands,
            EngineEvent::ControllerButtonUp(controller, button),
        )?;

        Ok(())
    }

    #[inline]
    fn button_down_event<T: EventConsumer>(
        commands: &mut Commands,
        gui_handler: &mut GuiHandler,
        consumer: &mut T,
        controller: &Controller,
        button: ControllerButton,
    ) -> Result<()> {
        if gui_handler.check_navigatable() {
            Self::check_button_down(commands, gui_handler, consumer, controller, button)?;
        } else {
            consumer.event(
                commands,
                EngineEvent::ControllerButtonDown(controller, button),
            )?;
        }

        Ok(())
    }

    #[inline]
    fn axis_event<T: EventConsumer>(
        commands: &mut Commands,
        consumer: &mut T,
        controller: &Controller,
    ) -> Result<()> {
        consumer.event(
            commands,
            EngineEvent::ControllerAxis(controller, controller.controller_axis()),
        )?;

        Ok(())
    }

    #[inline]
    fn check_button_down<T: EventConsumer>(
        commands: &mut Commands,
        gui_handler: &mut GuiHandler,
        consumer: &mut T,
        controller: &Controller,
        button: ControllerButton,
    ) -> Result<()> {
        match button {
            ControllerButton::A => {
                if gui_handler.accept_selection()? {
                    return Ok(());
                }
            }

            ControllerButton::B => {
                if gui_handler.decline_topgui(commands)? {
                    return Ok(());
                }
            }

            ControllerButton::RightButton => {
                if gui_handler.next_tab_topgui(commands, false)? {
                    return Ok(());
                }
            }

            ControllerButton::LeftButton => {
                if gui_handler.previous_tab_topgui(commands, false)? {
                    return Ok(());
                }
            }

            ControllerButton::RightTrigger => {
                if gui_handler.next_tab_topgui(commands, true)? {
                    return Ok(());
                }
            }

            ControllerButton::LeftTrigger => {
                if gui_handler.previous_tab_topgui(commands, true)? {
                    return Ok(());
                }
            }

            ControllerButton::DPadDown => {
                let selection_res = gui_handler.update_selection(GuiDirection::Down)?;
                gui_handler.update_selection(GuiDirection::None)?;

                if selection_res {
                    return Ok(());
                }
            }

            ControllerButton::DPadUp => {
                let selection_res = gui_handler.update_selection(GuiDirection::Up)?;
                gui_handler.update_selection(GuiDirection::None)?;

                if selection_res {
                    return Ok(());
                }
            }

            ControllerButton::DPadRight => {
                let selection_res = gui_handler.update_selection(GuiDirection::Right)?;
                gui_handler.update_selection(GuiDirection::None)?;

                if selection_res {
                    return Ok(());
                }
            }

            ControllerButton::DPadLeft => {
                let selection_res = gui_handler.update_selection(GuiDirection::Left)?;
                gui_handler.update_selection(GuiDirection::None)?;

                if selection_res {
                    return Ok(());
                }
            }

            _ => (),
        }

        if !gui_handler.accept_custom_selection(commands, button)? {
            consumer.event(
                commands,
                EngineEvent::ControllerButtonDown(controller, button),
            )?;
        }

        Ok(())
    }

    #[inline]
    fn text_input(gui_handler: &mut GuiHandler, text: String) -> Result<()> {
        if let Some(writeable) = gui_handler.writeable() {
            for c in text.chars() {
                writeable.add_letter(gui_handler, c)?;
            }
        }

        Ok(())
    }
}
