use crate::prelude::*;

use super::{
    RenderingCore,
    animator::*,
    raytraceshader::{DefaultHardwareRayTracingShader, RTInAWeekendShader},
};

use anyhow::Result;

use std::sync::{Arc, Mutex};

pub struct HardwareRayTracer {
    device: Arc<Device>,

    width: u32,
    height: u32,

    recursion_depth: u32,
    use_default_pipeline: bool,

    // required infos (tlas, target image, camera)
    ray_tracing_desc_set: Arc<DescriptorSet>,

    // pipeline handle
    ray_tracing_pipeline: Arc<Pipeline>,

    // shader binding table (nvidia magic)
    sbt: ShaderBindingTable,

    // acceleration structures
    blas: Option<Arc<AccelerationStructure>>,
    tlas: Option<Arc<AccelerationStructure>>,
}

impl HardwareRayTracer {
    fn create_ray_trace_descriptor_set(device: &Arc<Device>) -> Result<Arc<DescriptorSet>> {
        let desc_layout = DescriptorSetLayout::builder()
            // tlas
            .add_layout_binding(
                0,
                VK_DESCRIPTOR_TYPE_ACCELERATION_STRUCTURE_KHR,
                VK_SHADER_STAGE_RAYGEN_BIT_KHR | VK_SHADER_STAGE_CLOSEST_HIT_BIT_KHR,
                0,
            )
            // result image
            .add_layout_binding(
                1,
                VK_DESCRIPTOR_TYPE_STORAGE_IMAGE,
                VK_SHADER_STAGE_RAYGEN_BIT_KHR,
                0,
            )
            // depth buffer
            .add_layout_binding(
                2,
                VK_DESCRIPTOR_TYPE_STORAGE_IMAGE,
                VK_SHADER_STAGE_RAYGEN_BIT_KHR,
                0,
            )
            // camera
            .add_layout_binding(
                3,
                VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER,
                VK_SHADER_STAGE_RAYGEN_BIT_KHR,
                0,
            )
            .build(device.clone())?;

        let descriptor_pool = DescriptorPool::builder()
            .set_layout(desc_layout)
            .set_flags(VK_DESCRIPTOR_POOL_CREATE_UPDATE_AFTER_BIND_BIT_EXT)
            .build(device.clone())?;

        descriptor_pool.prepare_set().allocate()
    }

    fn create_pipeline(
        use_default_pipeline: bool,
        recursion_depth: u32,
        pipeline_layout: Arc<PipelineLayout>,
        device: &Arc<Device>,
    ) -> Result<(Arc<Pipeline>, ShaderBindingTable)> {
        if use_default_pipeline {
            let shader = DefaultHardwareRayTracingShader::new(device, recursion_depth)?;

            Pipeline::new_ray_tracing()
                .add_shader(shader.ray_gen, None, None)
                .add_shader(shader.miss, None, None)
                .add_shader(shader.shadow_miss, None, None)
                .add_hit_shaders(
                    vec![
                        (HitShader::from(shader.closest_hit), None),
                        (HitShader::from(shader.shadow_hit), None),
                    ],
                    None,
                )
                .max_recursion_depth(recursion_depth)
                .build(device.clone(), &pipeline_layout)
        } else {
            let shader = RTInAWeekendShader::new(device, recursion_depth)?;

            Pipeline::new_ray_tracing()
                .add_shader(shader.ray_gen, None, None)
                .add_shader(shader.miss, None, None)
                .add_hit_shaders(vec![(HitShader::from(shader.closest_hit), None)], None)
                .max_recursion_depth(recursion_depth)
                .build(device.clone(), &pipeline_layout)
        }
    }
}

impl RenderingCore for HardwareRayTracer {
    fn new(
        device: &Arc<Device>,
        _queue: &Arc<Mutex<Queue>>,
        ray_tracing_info: RaytracingInfo,
        view: &View,
        target_image: &Arc<Image>,
        depth_image: &Arc<Image>,
        descriptors: Vec<&dyn VkHandle<VkDescriptorSetLayout>>,
    ) -> Result<Self> {
        let ray_tracing_desc_set = Self::create_ray_trace_descriptor_set(device)?;

        ray_tracing_desc_set.update(&[
            DescriptorWrite::storage_images(1, &[target_image]),
            DescriptorWrite::storage_images(2, &[depth_image]),
            DescriptorWrite::uniform_buffers(3, &[view.buffer()]),
        ])?;

        let mut pipeline_layout_builder =
            PipelineLayout::builder().add_descriptor_set_layout(&ray_tracing_desc_set);

        for descriptor in descriptors {
            pipeline_layout_builder = pipeline_layout_builder.add_descriptor_set_layout(descriptor);
        }

        let pipeline_layout = pipeline_layout_builder.build(device.clone())?;

        let recursion_depth = device
            .physical_device()
            .ray_tracing_properties()
            .maxRayRecursionDepth;

        let (ray_tracing_pipeline, sbt) = Self::create_pipeline(
            ray_tracing_info.use_default_pipeline,
            recursion_depth,
            pipeline_layout,
            device,
        )?;

        Ok(HardwareRayTracer {
            device: device.clone(),

            recursion_depth,
            use_default_pipeline: ray_tracing_info.use_default_pipeline,

            width: target_image.width(),
            height: target_image.height(),

            ray_tracing_desc_set,

            ray_tracing_pipeline,

            sbt,

            blas: None,
            tlas: None,
        })
    }

    fn recompile(&mut self, descriptors: &[&dyn VkHandle<VkDescriptorSetLayout>]) -> Result<()> {
        let mut pipeline_layout_builder =
            PipelineLayout::builder().add_descriptor_set_layout(&self.ray_tracing_desc_set);

        for descriptor in descriptors {
            pipeline_layout_builder =
                pipeline_layout_builder.add_descriptor_set_layout(*descriptor);
        }

        let pipeline_layout = pipeline_layout_builder.build(self.device.clone())?;

        let (ray_tracing_pipeline, sbt) = Self::create_pipeline(
            self.use_default_pipeline,
            self.recursion_depth,
            pipeline_layout,
            &self.device,
        )?;

        self.ray_tracing_pipeline = ray_tracing_pipeline;
        self.sbt = sbt;

        Ok(())
    }

    fn check(device: &Arc<Device>) -> bool {
        if device
            .physical_device()
            .ray_tracing_features()
            .rayTracingPipeline
            == VK_FALSE
        {
            println!("ray tracing is not available!");

            return false;
        }

        if device
            .physical_device()
            .acceleration_structure_features()
            .accelerationStructure
            == VK_FALSE
        {
            println!("acceleration structure is not available!");

            return false;
        }

        let properties = device.physical_device().properties();

        if properties.limits.maxBoundDescriptorSets < 8 {
            println!("required descriptor set count not met");

            return false;
        }

        true
    }

    fn resize(&mut self, image: &Arc<Image>, positions: &Arc<Image>) -> Result<()> {
        self.width = image.width();
        self.height = image.height();

        self.ray_tracing_desc_set.update(&[
            DescriptorWrite::storage_images(1, &[image]),
            DescriptorWrite::storage_images(2, &[positions]),
        ])?;

        Ok(())
    }

    fn invalidate(&mut self) {
        self.blas = None;
        self.tlas = None;
    }

    fn process(
        &mut self,
        buffer_recorder: &mut CommandBufferRecorder<'_>,
        descriptors: &[&Arc<DescriptorSet>],
        scene_buffer: &Arc<Buffer<VertexData>>,
        _index: usize,
    ) -> Result<()> {
        // assume that tlas is_none always at the same as blas is
        if self.blas.is_none() {
            // create blas
            let blas = AccelerationStructure::bottom_level()
                .add_vertices(
                    scene_buffer,
                    None,
                    VK_GEOMETRY_NO_DUPLICATE_ANY_HIT_INVOCATION_BIT_KHR,
                )
                .build(self.device.clone(), buffer_recorder)?;

            // create tlas
            let tlas = AccelerationStructure::top_level()
                .add_instance(
                    &blas,
                    None,
                    VK_GEOMETRY_INSTANCE_TRIANGLE_FACING_CULL_DISABLE_BIT_KHR,
                )
                .build(self.device.clone(), buffer_recorder)?;

            // update AS descriptor with the newest top level AS
            self.ray_tracing_desc_set
                .update(&[DescriptorWrite::acceleration_structures(0, &[&tlas])])?;

            self.blas = Some(blas);
            self.tlas = Some(tlas);
        } else {
            self.blas.as_ref().unwrap().update(buffer_recorder)?;
            self.tlas.as_ref().unwrap().update(buffer_recorder)?;
        }

        let mut descs = Vec::with_capacity(descriptors.len() + 1);
        descs.push(&self.ray_tracing_desc_set);

        for descriptor in descriptors {
            descs.push(*descriptor);
        }

        // add ray tracing commands to the command buffer
        buffer_recorder.bind_pipeline(&self.ray_tracing_pipeline)?;
        buffer_recorder.bind_descriptor_sets_minimal(&descs);
        buffer_recorder.trace_rays(&self.sbt, self.width, self.height, 1);

        Ok(())
    }
}
