#version 450

#ifndef AMBIENT
    #define AMBIENT 0.2
#endif

#define M_PI 3.14159265358979323846264338327950288

// input parameter
layout (location = 0) in vec3 in_world_pos;

#ifdef COLOR_UVS
    layout (location = 1) in vec2 in_uv;
#endif

#ifdef COLOR
    layout (location = 1) in vec4 in_color;
#endif

#ifdef NORMALS
    #ifdef NORMAL_UVS
        layout (location = 2) in vec3 in_tangent;
    #endif

    layout (location = 3) in vec3 in_normal;
#endif

layout (set = 0, binding = 1) uniform Material {
    vec4 color;

    int color_texture_index;
    int normal_texture_index;
    int emissive_texture_index;
    int roughness_texture_index;

    float metallic_factor;
    vec3 emissive_factor;

    float roughness_factor;
} material;

// output parameter
layout (location = 0) out vec4 out_frag_color;
layout (location = 1) out vec4 out_positon;

// descriptors
#ifdef COLOR_UVS
	layout (set = 0, binding = 2) uniform sampler2D color_texture;
#endif

#ifdef NORMAL_UVS
	layout (set = 0, binding = 4) uniform sampler2D normal_texture;
#endif

#if POINT_LIGHT_COUNT > 0
    layout (set = 2, binding = 0) uniform PointLight {
        vec4 position;
        vec3 color;
        float power;
        mat4 view;
    } point_light_infos[POINT_LIGHT_COUNT];

    layout (set = 3, binding = 0) uniform sampler2D point_light_shadows[POINT_LIGHT_COUNT];
#endif // POINT_LIGHT_COUNT > 0

#if DIRECTIONAL_LIGHT_COUNT > 0
    #if POINT_LIGHT_COUNT == 0
        #define DIRECTIONAL_INDEX 2
    #else
        #define DIRECTIONAL_INDEX 4
    #endif

    layout (set = DIRECTIONAL_INDEX, binding = 0) uniform DirectionalLight {
        vec4 direction;
        vec3 color;
        float power;
        mat4 view;
        vec4 light_position;
    } directional_light_infos[DIRECTIONAL_LIGHT_COUNT];

    layout (set = DIRECTIONAL_INDEX + 1, binding = 0) uniform sampler2D directional_light_shadows[DIRECTIONAL_LIGHT_COUNT];
#endif // DIRECTIONAL_LIGHT_COUNT > 0

#if SPOT_LIGHT_COUNT > 0
    #if POINT_LIGHT_COUNT == 0 && DIRECTIONAL_LIGHT_COUNT == 0
        #define SPOT_INDEX 2
    #elif POINT_LIGHT_COUNT == 0 || DIRECTIONAL_LIGHT_COUNT == 0
        #define SPOT_INDEX 4
    #else
        #define SPOT_INDEX 6
    #endif

    layout (set = SPOT_INDEX, binding = 0) uniform SpotLight {
        vec4 position;
        vec3 direction;
        float angle;
        vec3 color;
        float power;
        mat4 view;
    } spot_light_infos[SPOT_LIGHT_COUNT];

    layout (set = SPOT_INDEX + 1, binding = 0) uniform sampler2D spot_light_shadows[SPOT_LIGHT_COUNT];
#endif // SPOT_LIGHT_COUNT > 0

bool texture_proj(vec4 P, sampler2D tex)
{
	vec4 shadow_coord = P / P.w;
	shadow_coord.st = shadow_coord.st * 0.5 + 0.5;

	if (shadow_coord.z > -1.0 && shadow_coord.z < 1.0)
	{
		float dist = texture(tex, shadow_coord.st).r;

		if (shadow_coord.w > 0.0 && dist < shadow_coord.z)
		{
			return false;
		}
	}

	return true;
}

float rescale(float x, float x0, float x1, float y0, float y1) {
    return y0 + (y1 - y0) * (x - x0) / (x1 - x0);
}

vec3 rescale(vec3 v, float x0, float x1, float y0, float y1) {
    return vec3(
        rescale(v.x, x0, x1, y0, y1),
        rescale(v.y, x0, x1, y0, y1),
        rescale(v.z, x0, x1, y0, y1)
    );
}

float intensity(float power, float distance) {
    if (distance == 0.0) {
        return 1.0;
    }

    // surface of a sphere
    // A = 4 * PI * r^2
    float surface = 4 * M_PI * distance * distance;

    return min(1.0, power / surface);
}


void main()
{
    vec3 light_factor = vec3(0.0);

    #ifdef NORMALS
        vec3 normal = normalize(in_normal);

        #ifdef NORMAL_UVS
            normal.y = -normal.y;

            vec3 T = normalize(in_tangent);
            vec3 B = cross(normal, T);
            mat3 TBN = mat3(T, B, normal);

            normal = TBN * normalize(texture(normal_texture, in_uv).xyz * 2.0 - vec3(1.0));
        #endif

        #if DIRECTIONAL_LIGHT_COUNT > 0
            for (int i = 0; i < DIRECTIONAL_LIGHT_COUNT; i++) {
                vec3 direction = directional_light_infos[i].light_position.xyz - in_world_pos;
                float dir_length = length(direction);
                direction /= dir_length;

                // TODO
                // float intensity = intensity(directional_light_infos[i].power, dir_length);
                float intensity = 1.0;

                float n_dot_l = dot(direction, normal);

                float factor = 0.0;

                #ifdef SHADOW_MAPS
                    if (intensity > 0.03 && n_dot_l > 0.0) {
                        vec4 shadow_clip = directional_light_infos[i].view * vec4(in_world_pos, 1.0);
                        bool shadowed = texture_proj(shadow_clip, directional_light_shadows[i]);

                        if (!shadowed) {
                            factor = n_dot_l * intensity;
                        }
                    }
                #else
                    factor = clamp(n_dot_l, 0.0, 1.0) * intensity;
                #endif

                light_factor = max(light_factor, directional_light_infos[i].color * factor);
            }
        #endif // DIRECTIONAL_LIGHT_COUNT

        #if POINT_LIGHT_COUNT > 0
            for (int i = 0; i < POINT_LIGHT_COUNT; i++) {
                vec3 direction = point_light_infos[i].position.xyz - in_world_pos;
                float dir_length = length(direction);
                direction /= dir_length;

                // TODO
                float intensity = intensity(point_light_infos[i].power, dir_length);
                // float intensity = 1.0;

                float n_dot_l = dot(direction, normal);

                float factor = 0.0;

                #ifdef SHADOW_MAPS
                    if (intensity > 0.03 && n_dot_l > 0.0) {
                        // TODO

                        // vec4 shadow_clip = point_light_infos[i].view * vec4(in_world_pos, 1.0);
                        // bool shadowed = texture_proj(shadow_clip, point_light_shadows[i]);

                        // if (!shadowed) {
                        //     factor = n_dot_l * intensity;
                        // }
                    }
                #else
                    factor = clamp(n_dot_l, 0.0, 1.0) * intensity;
                #endif

                light_factor = max(light_factor, point_light_infos[i].color * factor);
            }
        #endif // POINT_LIGHT_COUNT

        // TODO: spot light
    #endif

    #ifdef COLOR_UVS
        vec4 albedo = material.color * texture(color_texture, in_uv);
    #else
        #ifdef COLOR
            vec4 albedo = in_color;
        #else
            vec4 albedo = material.color;
        #endif
    #endif

    vec3 color = albedo.xyz * rescale(light_factor, 0.0, 1.0, AMBIENT, 1.0);

    out_frag_color = vec4(color, albedo.w);
    out_positon = vec4(in_world_pos, 1.0);
}
