use context::prelude::*;
use ecs::Commands;

use std::{
    sync::{
        Arc,
        mpsc::{Receiver, channel},
    },
    thread,
};

use anyhow::{Result, anyhow};

pub struct LoadingScreen<T, G>
where
    T: Send + Sync + 'static,
    G: TopLevelGui + TopGui + Send + Sync + 'static,
{
    result: Option<T>,
    receiver: Receiver<T>,
    gui: Arc<G>,
}

impl<T, G> LoadingScreen<T, G>
where
    T: Send + Sync + 'static,
    G: TopLevelGui + TopGui + Send + Sync + 'static,
{
    pub fn load<R, L>(
        commands: &mut Commands,
        gui_handler: &mut GuiHandler,
        gui: Arc<G>,
        loader: L,
    ) -> Result<Self>
    where
        R: Fn(T) -> Result<()> + Send + Sync + 'static,
        T: Send + Sync + 'static,
        L: FnOnce() -> T + Send + Sync + 'static,
    {
        gui.enable(commands, gui_handler)?;

        let (sender, receiver) = channel();

        thread::spawn(move || {
            let _ = sender.send(loader());
        });

        Ok(Self {
            result: None,
            receiver,
            gui,
        })
    }

    pub fn check_ready(&mut self) -> bool {
        if self.result.is_some() {
            return true;
        }

        match self.receiver.try_recv() {
            Ok(result) => {
                self.result = Some(result);
                true
            }
            Err(_) => false,
        }
    }

    pub fn take_result(
        mut self,
        commands: &mut Commands,
        gui_handler: &mut GuiHandler,
    ) -> Result<T> {
        self.gui.disable(commands, gui_handler)?;
        self.result
            .take()
            .ok_or(anyhow!("missing loading screen payload!"))
    }
}
